use std::{fmt::Debug, sync::Arc, time::Duration};

use domain::entities::account::Account;
use lazy_static::lazy_static;
use library::config;
use moka::{
    future::{Cache, CacheBuilder}, notification::RemovalCause, policy::EvictionPolicy
};
use serde::{de::DeserializeOwned, Deserialize, Serialize};

use crate::core::redis::REDIS_CONN;

#[derive(Debug, Clone, Serialize, Deserialize)]
pub struct CacheAccount {
    pub account: Arc<Account>,
    pub token: Arc<String>,
}

pub struct ServerCache<S, D> {
    pub cache: Cache<S, D>,
}

impl<S, D> Default for ServerCache<S, D>
where
    S: std::hash::Hash + Eq + Debug + ToString + Send + Sync + 'static,
    D: Debug + Send + Sync + Clone + Serialize + DeserializeOwned + 'static,
{
    fn default() -> Self {
        ServerCache {
            cache: CacheBuilder::new(20480)
            .name("login_cache")
            .eviction_policy(EvictionPolicy::lru())
            .time_to_live(Duration::from_secs(config!().jwt.expires as u64))
            .async_eviction_listener(|key: Arc<S>, value: D, cause| Box::pin(async move {
                tracing::info!(
                    "内存缓存数据移除 key: {:?}, value: {:?}, cause: {:?}",
                    key,
                    value,
                    cause
                );
                if cause != RemovalCause::Replaced {
                    RedisCache {
                        cache_type: std::any::type_name::<D>().to_string(),
                        cache_key: key.to_string(),
                        ..Default::default()
                    }.remove().await;
                }
            }))
            .build(),
        }
    }
}

impl<D> ServerCache<String, D>
where
    D: Debug + Send + Sync + Clone + Serialize + DeserializeOwned + 'static,
{
    pub async fn init(&self) {
        let cache_type = std::any::type_name::<D>().to_string();
        let cache_datas = RedisCache::get(cache_type).await;
        for element in cache_datas {
            let cache_key = element.cache_key.clone();
            let data_json = element.data;
            let cache_data = serde_json::from_str::<D>(&data_json).unwrap();
            self.add_cache(cache_key, cache_data).await;
        }
    }

    async fn add_cache(&self, key: String, value: D) {
        self.cache.insert(key, value).await;
    }

    pub async fn insert(&self, key: String, value: D) {
        let cache_data = RedisCache{
            cache_type: std::any::type_name::<D>().to_string(), 
            data: serde_json::to_string(&value).unwrap(),
            cache_key: key.to_string(),
            ..Default::default()
        };
        cache_data.insert().await;
        self.cache.insert(key, value).await;
    }

    pub async fn get(&self, key: &str) -> Option<D> {
        self.cache.get(key).await
    }

    pub async fn remove(&self, key: &str) {
        self.cache.remove(key).await;
    }
}

#[derive(Debug, Clone, Default)]
struct RedisCache {
    pub cache_type: String,
    pub cache_key: String,
    pub data: String,
}

impl RedisCache {
    pub async fn insert(&self) {
        match REDIS_CONN.hset(&self.cache_type, &self.cache_key, &self.data).await {
            Ok(_) => {},
            Err(err) => {
                tracing::error!("向缓存Set插入数据失败: {:?}", err);
            }
        }
    }

    pub async fn get(cache_type: String) -> Vec<RedisCache> {
        let mut result = Vec::new();
        match REDIS_CONN.hgetall(&cache_type).await {
            Ok(cache_datas) => {
                for (key, value) in cache_datas {
                    let cache_data = RedisCache {
                        cache_type: cache_type.clone(),
                        cache_key: key,
                        data: value,
                    };
                    result.push(cache_data);
                }
            },
            Err(err) => {
                tracing::error!("从缓存Set获取数据失败: {:?}", err);
            },
        }
        
        result
    }

    pub async fn remove(&self) {
        match REDIS_CONN.hdel(&self.cache_type, &self.cache_key).await {
            Ok(_) => {},
            Err(err) => {
                tracing::error!("从缓存Set删除数据失败: {:?}", err);
            }
        }
    }
}

// impl CacheData {
//     /// 向缓存数据表中插入数据
//     /// 
//     /// 先判断数据是否存在，若已经存在，则删除
//     /// 
//     /// 然后保存新的缓存数据
//     pub async fn insert(&self) {
//         let mut transaction = db!().begin().await.unwrap();
//         let exist_cache_data = match sqlx::query_as!(
//             CacheData, 
//             r#"select * from cache_data where cache_key = $1"#, 
//             self.cache_key
//         ).fetch_all(&mut *transaction).await {
//             Ok(data) => data,
//             Err(err) => {
//                 tracing::error!("从数据库获取缓存数据失败: {:?}", err);
//                 tracing::error!("{}", err);
//                 Vec::new()
//             }
//         };
//         for ele in exist_cache_data {
//             match sqlx::query!(
//                 r#"
//                     delete from cache_data where id = $1
//                 "#,
//                 ele.id
//             )
//             .execute(&mut *transaction)
//             .await {
//                 Ok(_) => {},
//                 Err(err) => {
//                     tracing::error!("从数据库删除缓存数据失败: {:?}", err);
//                     tracing::error!("{}", err);
//                     transaction.rollback().await.unwrap();
//                     return ;
//                 }
//             }
//         }

//         match sqlx::query!(
//             r#"
//                 insert into cache_data (cache_type, cache_key, data)
//                 values ($1, $2, $3)
//             "#,
//             self.cache_type,
//             self.cache_key,
//             self.data
//         )
//         .execute(&mut *transaction)
//         .await
//         {
//             Ok(_) => transaction.commit().await.unwrap(),
//             Err(err) => {
//                 tracing::error!("向数据库插入缓存数据失败: {:?}", err);
//                 transaction.rollback().await.unwrap()
//             },
//         }
//     }

//     pub async fn get(cache_type: String) -> Vec<CacheData> {
//         let mut cache_data = Vec::new();
//         match sqlx::query_as!(
//             CacheData,
//             r#"
//                 select id, cache_type, cache_key, data from cache_data where cache_type = $1
//             "#,
//             cache_type
//         )
//         .fetch_all(db!())
//         .await
//         {
//             Ok(data) => {
//                 cache_data = data;
//             }
//             Err(err) => {
//                 tracing::error!("从数据库获取缓存数据失败: {:?}", err);
//                 tracing::error!("{}", err);
//             }
//         }
//         cache_data
//      }

//      pub async fn remove(cache_key: String) {
//         match sqlx::query!(
//             r#"
//                 delete from cache_data where cache_key = $1
//             "#,
//             cache_key
//         )
//         .execute(db!())
//         .await
//         {
//             Ok(_) => {},
//             Err(err) => {
//                 tracing::error!("从数据库删除缓存数据失败: {:?}", err);
//                 tracing::error!("{}", err);
//             }
//         }
//      }
// }

pub async fn init_cache() {
    LOGIN_ACCOUNT_CACHE.init().await;
}

lazy_static! {
    pub static ref LOGIN_ACCOUNT_CACHE: ServerCache<String, Arc<CacheAccount>> = ServerCache::default();
}
